#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import numpy as np


def get_bracketing_indexes(array, value):
    """Returns consecutive indexes idx_low and idx_high such that array[idx_low] <= value < array[idx_high].

    :param array: Array to search. Must be 1D and monotonically sorted in ascending order. This condition is NOT checked here
    :type array: array_like

    :param value: Value to bracket
    :type value: Union[float, np.datetime64, np.timedelta64]

    :returns: Indexes that fulfill the condition array[idx_low] <= value < array[idx_high].
    :rtype: tuple

    """

    i_high = np.argmin(array <= value)
    i_low = i_high - 1
    return i_low, i_high



# -*- coding: utf-8 -*-

import xarray as xr
import itertools
from metintos.optiflow import MultiFrameInterpolant
from metintos.utils import *
from math import ceil


def check_array_for_nans(arr):
    """

    :param arr:
    :type arr:

    :returns: None

    """

    assert not np.isnan(arr.sum())


class CoordinateGenerator(object):
    """Helper class for generating coordinates

        :var dict: Coordinates dictionary, mapping axis_name -> np.ndarray

    """

    def __init__(self):
        """This declares the axes variable as a dictionary """

        self.axes = {}

    def add_axis_lims_resolution(self, ax_name, lower, upper, resolution):
        """Add an axis coordinate specifying limits and resolution (i.e. steps between coordinate points)

            :param ax_name: Name of the coordinate axis to be set
            :type ax_name: str

            :param lower: Lower limit of the coordinate axis
            :type lower: float

            :param upper: Upper limit of the coordinate axis
            :type upper: float

            :param resolution: Distance between axis points
            :type resolution: float

            :returns: None

        """
        self.axes[ax_name] = np.arange(lower, upper, resolution)

    def add_axis_lims_n_points(self, ax_name, lower, upper, n, dtype=None):
        """Add an axis coordinate specifying limits and number of points in coordinate axis

            :param ax_name: Name of the coordinate axis to be set
            :type ax_name: str

            :param lower: Lower limit of the coordinate axis
            :type lower: float

            :param upper: Upper limit of the coordinate axis
            :type upper: float

            :param n: Number of points in the coordinate axis
            :type n: int

            :param dtype: The type of the coordinate array. If dtype is not given, it will be inferred by numpy.linspace()
            :type dtype: dtype, optional

            :returns: None

        """
        #self.axes[ax_name] = np.linspace(lower, upper, n, dtype=dtype)
        self.axes[ax_name] = lower + (upper - lower) * np.linspace(0, 1, n)


def get_var_name(dataset, variable):
    """

        :param dataset:
        :type dataset:

        :param variable:
        :type variable:

        :returns: None

    """
    try:
        return dataset[variable].shortname
    except AttributeError:
        return variable


class DatasetHandler(object):
    """Wrapper for a xarray.Dataset containing meteorological information at pressure levels

        :var ds: The dataset
        :type ds: xarray.Dataset


    """

    def __init__(self, ds, output_dtype=np.float32, time_dim='step'):
        """
            :param ds: The dataset
            :type ds: xarray.Dataset

            :param output_dtype: (optional)
            :type output_dtype: data-type

            :param time_dim: (optional)
            :type time_dim: str

            :returns: None

        """
        self.time_dim = time_dim
        self.ds = ds
        self.output_dtype = np.float32
        lat = self.ds.coords['latitude']
        lon = self.ds.coords['longitude']

        # check if we are under a dataset with ensemble
        self.is_ensemble = False
        try:
            self.is_ensemble = len(self.ds.number.values)
        except TypeError:
            pass

        if len(lat.shape) == 1:
            self.geo_grid = 'latlon'
            if lat[0] > lat[1]:
                self.ds = self.ds.reindex(latitude=lat[::-1])
            # Monotonically ascending coordinates are required by some interpolants
        elif len(lat.shape) == 2:
            self.geo_grid = 'xy'
            if lat[0, 0] > lat[1, 0]:
                self.ds = self.ds.reindex(y=self.ds.coords['y'][::-1])
            # Monotonically ascending coordinates are required by some interpolants
        else:
            raise ValueError(f"Cannot handle lat/lon coordinates with shape {lat.shape}")
        self.ds = self.ds.transpose(*self.default_dims)

    @classmethod
    def load_from_steps(cls, path_list, **backend_kwargs):
        """Loads the meteo infromation from multiple grib files with consecutive forecast steps

            :param path_list: List of paths of the grib files
            :type path_list: list

            :param backend_kwargs:
            :type backend_kwargs:

            :returns: cls(ds)

        """
        ds = xr.open_mfdataset(path_list, engine='cfgrib', concat_dim=['step'], combine='nested', **backend_kwargs)
        return cls(ds)

    @property
    def default_dims(self):
        """Returns the default dimensions

            :returns: tuple with the dimensions.

        """
        if self.geo_grid == 'latlon':
            dim_list = ('number', 'isobaricInhPa', self.time_dim, 'latitude', 'longitude')
        else:
            dim_list = ('number', 'isobaricInhPa', self.time_dim, 'y', 'x')
        return tuple(dim for dim in dim_list if dim in self.ds.dims.keys())

    @ property
    def data_variables_list(self):
        """Return a list of variables that are not coordinates

            :returns: list of variables that are not coordinates

        """
        all_vars = set(self.ds.variables.keys())
        coords = set(self.ds.coords)
        data_vars = all_vars - coords
        return data_vars

    def get_geographical_coordinate_slice_by_indexes(self, lat_idx_low, lat_idx_high, lon_idx_low, lon_idx_high):
        """Get a CoordinateGenerator (Not implemented)

            :param lat_idx_low:
            :type lat_idx_low:

            :param lat_idx_high:
            :type lat_idx_high:

            :param lon_idx_low:
            :type lon_idx_low:

            :param lon_idx_high:
            :type lon_idx_high:

            :returns: None

        """

    def complete_new_coords(self, **coords):
        """Generates a coordinate set by completing the specificed new coordinates with the already existing ones


            :param coords: New coordinates, as a dict mapping variable names to monotonically increasing 1D arrays
            :type coords: dict

            :returns: New coordinates, completing missing dimensions with the already existing values
            :rtype: dict

        """
        new_coords = {}
        for ax_name, ax in self.ds.coords.items():
            new_coords[ax_name] = ax
        for ax_name, ax in coords.items():
            new_coords[ax_name] = xr.IndexVariable(ax_name, ax, attrs=self.ds.coords[ax_name].attrs)
        try:
            del new_coords['valid_time']
        except KeyError:
            pass
        return new_coords

    def get_resampled_dataset(self, **coords):
        """Resamples the dataset to new coordinates


            :param coords: A dictionary of new coordinates, mapping coordinate names to numpy arrays. If a coordinate axis is missing, the old coordinate axis is used by default
            :type coords: dict

            :returns: A dataset reinterpolated to the new coordinates
            :rtype: xarray.Dataset

        """
        data_vars = self.data_variables_list
        for dv in data_vars:  # fake loop to pick any element from the data_vars set
            break
        dims = self.ds[dv].dims
        new_coords = self.complete_new_coords(**coords)
        new_shape = tuple(len(new_coords[d]) for d in dims)
        placeholder_dataarray = xr.DataArray(data=np.zeros(new_shape), dims=dims, coords=new_coords)
        return self.ds.interp_like(placeholder_dataarray)

    def get_optical_flow_interpolated_dataset(self, new_coords, variables=None, flow_calculator=None,
                                              spatial_interp=None, verbose=False, zero_out_nans=True):
        """Reinterpolates the dataset to the new coordinates using optical flow interpolation


            :param new_coords: A dictionary of new coordinates, mapping coordinate names to numpy arrays. If a coordinate axis is missing, the old coordinate axis is used by default
            :type new_coords: dict

            :param variables: (optional) List of variables to be transported to the new dataset
            :type variables: list

            :param flow_calculator: (optional) Callable that computes the optical flow between two frames / arrays
            :type flow_calculator: (R² x R²) -> R² mapping

            :param spatial_interp: (optional) Callable that generates a 2D interpolant of a frame array with normalized coordinates
            :type spatial_interp: R² -> (R² -> R)

            :param verbose: (optional) If true, display the progress of the dataset reinterpolation procedure
            :type verbose: bool

            :param zero_out_nans: (optional) if True, replace any nans in the input with zeros
            :type zero_out_nans: bool


            :returns: The reinterpolated dataset
            :rtype: xarray.Dataset


        """
        new_coords = self.complete_new_coords(**new_coords)
        if variables is None:
            variables = [get_var_name(self.ds, var) for var in self.ds.data_vars]
        new_vars = {var: self.get_optical_flow_interpolated_variable(new_coords, var, flow_calculator, spatial_interp,
                                                                     verbose=verbose, zero_out_nans=zero_out_nans)
                    for var in variables}
        return xr.Dataset(data_vars=new_vars, coords=new_coords, attrs=self.ds.attrs)

    @staticmethod
    def get_coords_shape(dims, coords):
        """Returns the shape of an array with the given coordinates and dimension ordering

            :param dims: dimension
            :type dims:

            :param coords: coordinates
            :type coords:


            :returns: The coordinates shape
            :rtype: tuple
        """
        ll_dims = len(coords['latitude'].shape)
        if ll_dims == 1:
            return tuple(coords[d].size for d in dims)
        elif ll_dims == 2:
            geo_coords = ('y', 'x')
            return tuple(coords[d].size for d in dims if d not in geo_coords) + coords['latitude'].shape

    def get_optical_flow_interpolated_variable(self, new_coords, variable,
                                               flow_calculator=None, spatial_interp=None,
                                               flow_padding=.1, verbose=False, zero_out_nans=True):
        """

            :param new_coords: A dictionary of new coordinates, mapping coordinate names to numpy arrays. If a coordinate axis is missing, the old coordinate axis is used by default
            :type new_coords: dict

            :param variable: variable to be transported to the new dataset
            :type variable: str

            :param flow_calculator: (optional) Callable that computes the optical flow between two frames / arrays
            :type flow_calculator: (R² x R²) -> R² mapping

            :param spatial_interp: (optional) Callable that generates a 2D interpolant of a frame array with normalized coordinates
            :type spatial_interp: R² -> (R² -> R)

            :param flow_padding: (optional)
            :type flow_padding: float

            :param verbose: (optional) If true, display the progress of the dataset reinterpolation procedure
            :type verbose: bool

            :param zero_out_nans: (optional) if True, replace any nans in the input with zeros
            :type zero_out_nans: bool


            :returns: The reinterpolated dataset
            :rtype: xarray.Dataset

        """

        reinterpolate = False
        reinterpolate_ens_level = False
        old_coords = self.ds.coords
        intermediate_coords = {}
        if self.time_dim in new_coords:
            intermediate_coords[self.time_dim] = new_coords[self.time_dim]
        limits = {'latitude': (-90.0, 90.0), 'longitude': (-180.0, 180.0)}
        geo_coords = limits.keys()
        for coord in geo_coords:
            try:
                assert np.allclose(new_coords[coord].values, old_coords[coord].values)
            except (ValueError, AssertionError):
                reinterpolate = self.geo_grid

        icg = CoordinateGenerator()
        if reinterpolate == "latlon":
            for coord in geo_coords:
                if coord in new_coords:
                    axis_new = new_coords[coord].values
                    axis_old = old_coords[coord].values
                    resolution = axis_old[1] - axis_old[0]  # assuming regular lat/lon grids
                    span = axis_new[-1] - axis_new[0]
                    padding = flow_padding * span
                    padding_indexes = ceil(padding / resolution)
                    low = max(axis_new[0] - padding_indexes * resolution, limits[coord][0], axis_old.min())
                    high = min(axis_new[-1] + (padding_indexes + 1) * resolution, limits[coord][1], axis_old.max())
                    icg.add_axis_lims_resolution(coord, low, high, resolution)
            intermediate_coords.update(icg.axes)
            reinterpolate_ens_level = True
        elif reinterpolate == "xy":
            raise NotImplementedError("Automatic clipping for optical flow calculation is not implemented for "
                                      "non-Plate-Carrée grids (i.e. 2D lat/lon grids). Regrid before or after"
                                      "optical flow resampling")
        else:
            icg.axes['latitude'] = old_coords['latitude']
            icg.axes['longitude'] = old_coords['longitude']
            for dim in self.default_dims:
                if dim not in (self.time_dim, ):
                    if dim in new_coords.keys():
                        try:
                            assert np.allclose(new_coords[dim].values, old_coords[dim].values)
                        except (ValueError, AssertionError):
                            reinterpolate_ens_level = True
                            break
        intermediate_coords = self.complete_new_coords(**intermediate_coords)
        intermediate_shape = self.get_coords_shape(self.default_dims, intermediate_coords)
        intermediate_var_array = np.zeros(intermediate_shape, dtype=self.output_dtype)
        intermediate_data_array = xr.DataArray(data=intermediate_var_array, coords=intermediate_coords,
                                               dims=self.default_dims, name=variable, attrs=self.ds[variable].attrs)
        coords_to_iterate = {}
        if self.is_ensemble:
            coords_to_iterate_keys = ('number', 'isobaricInhPa')
        else:
            coords_to_iterate_keys = ('isobaricInhPa',)
        for dim_key in coords_to_iterate_keys:
            try:
                coords_to_iterate[dim_key] = list(new_coords[dim_key].values)
            except KeyError:
                coords_to_iterate[dim_key] = [None]
        iterate_combinations = itertools.product(*coords_to_iterate.values())
        for coords_combi in iterate_combinations:
            if verbose:
                print(coords_combi)
            coords_selector_iter = {key: coords_combi[k] for k, key in enumerate(coords_to_iterate_keys)
                                    if coords_combi[k] is not None}
            coords_selector = coords_selector_iter.copy()
            if reinterpolate == 'latlon':
                coords_selector['latitude'] = icg.axes['latitude']
                coords_selector['longitude'] = icg.axes['longitude']

            def get_values(step, cs=coords_selector):
                _cs = cs.copy()
                _cs[self.time_dim] = step
                return self.ds[variable].sel(**_cs).values

            snapshots = [get_values(step)
                         for step in self.ds.coords[self.time_dim]]
            if zero_out_nans:
                snapshots = list(map(np.nan_to_num, snapshots))
            mfi = MultiFrameInterpolant(snapshots, flow_calculator, spatial_interp, old_coords[self.time_dim].values)
            for i, step in enumerate(new_coords[self.time_dim]):
                frame = mfi.eval_at_t(step.values)
                _csi = coords_selector_iter.copy()
                _csi[self.time_dim] = step
                intermediate_data_array.loc[dict(**_csi)] = frame
        if reinterpolate_ens_level:
            new_shape = self.get_coords_shape(self.default_dims, new_coords)
            new_var_array = np.zeros(new_shape, dtype=self.output_dtype)
            new_data_array = xr.DataArray(data=new_var_array,
                                          coords=new_coords,
                                          dims=self.default_dims,
                                          name=variable,
                                          attrs=self.ds[variable].attrs)
            return intermediate_data_array.interp_like(new_data_array)
        else:
            return intermediate_data_array



#!/usr/bin/python
# -*- coding: utf-8 -*-

from metintos.constants import *
from casadi import interpolant


class CasadiISA(object):
    """
    Casadi implementation of the ISA atmosphere (troposphere and tropopause only)

        :var deltaT: Temperature differential w.r.t. ISA conditions
        :type deltaT: float

    """

    def __init__(self, deltaT=0):
        """

            :param deltaT: Temperature differential w.r.t. ISA conditions
            :type deltaT: float, optional

            :returns: None

        """
        self.deltaT = deltaT

        z = np.arange(0, 20.5e3, 500)
        T = np.maximum(T0_MSL + lapse_rate_troposphere * z, T_tropopause * np.ones_like(z))

        P_tropo = P0_MSL * (T / T0_MSL) ** (-g / Rg / lapse_rate_troposphere)
        P_pause = P_tropo[22] * np.exp(-g / Rg / T * (z - h_tropopause))
        P = np.hstack([P_tropo[:23], P_pause[23:]])
        z_I = np.arange(0, 20.5e3, 341)
        T_I = np.maximum(T0_MSL + lapse_rate_troposphere * z_I, T_tropopause * np.ones_like(z_I))

        self.IP = interpolant('P', 'bspline', [z], P, {})
        self.IT = interpolant('T', 'bspline', [z_I], T_I, {})




import unittest
import xarray as xr
import metintos
import metintos.io
import numpy as np
import os


class TestDatasetHandler(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Set up data for the whole TestCase
        cwd = os.path.dirname(__file__)
        files_path = os.path.join(cwd, 'samples')
        files = [os.path.join(files_path, 'peticion_METATS_CONVECTION_Spain_2018-01-01_00_000_sfc.grib'),
                 os.path.join(files_path, 'peticion_METATS_CONVECTION_Spain_2018-01-01_00_001_sfc.grib'),
                 os.path.join(files_path, 'peticion_METATS_CONVECTION_Spain_2018-01-01_00_002_sfc.grib')]
        cls.ds = xr.open_mfdataset(files, engine='h5netcdf', concat_dim=['step'], combine='nested')

        # create DasetHandler object
        cls.dsh = metintos.io.DatasetHandler(cls.ds)

        # create CoordinateGenerator object and new step axi
        mn = np.timedelta64(1, 'm').astype('timedelta64[ns]')
        s0 = 0 * mn
        sf = 5 * 24.1 * mn
        ss = mn * 5
        cls.cg = metintos.io.CoordinateGenerator()
        # cg.add_axis_lims_n_points('latitude', 30, 70, 32)
        # cg.add_axis_lims_n_points('longitude', -10, 40, 32)
        cls.cg.add_axis_lims_resolution('step', s0, sf, ss)

        # produce new dataset with optical flow
        cls.dsn = cls.dsh.get_optical_flow_interpolated_dataset(cls.cg.axes, ['cape'])

    def test_dataset(self):
        # check ds
        list_variables = list(self.ds.variables.keys())
        self.assertEqual(len(list_variables), 22, 'Variables ds from xr.open_mfdataset are not equal')
        list_coords = list(self.ds.coords.keys())
        self.assertEqual(len(list_coords), 7, 'Coordinates ds from xr.open_mfdataset are not equal')
        list_dims = list(self.ds.dims.keys())
        self.assertEqual(len(list_dims), 4, 'Dimensions ds from xr.open_mfdataset are not equal')

    def test_datasetHandler(self):
        # check ds_copied is like ds
        ds_copied = self.dsh.ds
        list_variables = list(ds_copied.variables.keys())
        self.assertEqual(len(list_variables), 22, 'Variables ds in dsh from xr.open_mfdataset are not equal')
        list_coords = list(ds_copied.coords.keys())
        self.assertEqual(len(list_coords), 7, 'Coordinates ds in dsh  from xr.open_mfdataset are not equal')
        list_dims = list(ds_copied.dims.keys())
        self.assertEqual(len(list_dims), 4, 'Dimensions ds in dsh  from xr.open_mfdataset are not equal')

        # check default_dims
        default_dims = self.dsh.default_dims
        self.assertEqual(len(default_dims), 4, 'default dims in dsh are not equal')

        # check time_dim
        time_dim = self.dsh.time_dim
        self.assertEqual(time_dim, 'step', 'time dims in dsh are not equal')

    def test_coordinates_change(self):
        list_steps = self.cg.axes['step']
        self.assertEqual(len(list_steps), 25, 'New Steps axes to compute the optical flow are not equal')

    def test_optical_flow(self):
        list_variables = list(self.dsn.variables.keys())
        self.assertEqual(len(list_variables), 7, 'Variables dsn from xr.open_mfdataset are not equal')
        list_coords = list(self.dsn.coords.keys())
        self.assertEqual(len(list_coords), 6, 'Coordinates dsn from xr.open_mfdataset are not equal')
        list_dims = list(self.dsn.dims.keys())
        self.assertEqual(len(list_dims), 4, 'Dimensions dsn from xr.open_mfdataset are not equal')

        steps_of = self.dsn['step'].values
        self.assertEqual(len(steps_of), 25, 'Number of Optical flow steps are not the expected ones')
        self.assertEqual(min(steps_of).astype(int), 0, 'Min of Optical flow step is not 0')
        self.assertEqual(max(steps_of).astype(int), 7200000000000, 'Max of Optical flow step is not equal')

        cape_exaple_of = self.dsn.cape.values[0, 1, 10, 20]
        self.assertEqual(round(float(cape_exaple_of), 9), -0.001532894,
                         'cp value for the member=0, step=5min, latitude=XX, and longitude=XXX')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import cv2
from scipy.interpolate import RectBivariateSpline
from metintos.utils import *


class NormalizedWrappedInterpolant(object):
    """2D interpolant with natural index coordinates

        :var array: 2D array to interpolate
        :type array: array_like

        :var i: normalized coordinates along the first axis
        :type i: array_like

        :var j: normalized coordinates along the second axis
        :type j: array_like

        :var interpolant: wrapped interpolant
        :type interpolant: R² -> R mapping

    """

    def __init__(self, array, interpolant_generator=None):
        """

            :param array: 2D array to interpolate
            :type array: array_like

            :param interpolant_generator: wrapped interpolant
            :type interpolant_generator: R² -> (R² -> R) mapping

            :returns: None

        """

        if interpolant_generator is None:
            interpolant_generator = RectBivariateSpline

        self.shape = array.shape
        self.array = array
        self.i = np.linspace(0, self.shape[0] - 1, self.shape[0])
        self.j = np.linspace(0, self.shape[1] - 1, self.shape[1])
        self.interpolant = interpolant_generator(self.i, self.j, array)

    def __call__(self, *args, **kwargs):
        """

            :param args:
            :type args:

            :param kwargs:
            :type kwargs:

            :returns:

        """

        return self.interpolant(*args, **kwargs)

    def ev(self, *args, **kwargs):
        """

            :param args:
            :type args:

            :param kwargs:
            :type kwargs:

            :returns:

        """

        return self.interpolant.ev(*args, **kwargs)


class TwoFrameInterpolant(object):
    """Performs optical flow or linear interpolation between two consecutive frames

        :var start_frame: Matrix representation of the variable at the start of the temporal interpolation domain
        :type start_frame: 2D array

        :var end_frame: Matrix representation of the variable at the end of the temporal interpolation
        :type end_frame: 2D array

        :var flow_calculator: Callable that computes the optical flow between two frames / arrays
        :type flow_calculator: (R² x R²) -> R² mapping

        :var spatial_interp: Callable that generates a 2D interpolant of a frame array with normalized coordinates
        :type spatial_interp: R² -> (R² -> R)

        :var flow: Optical flow between the start and end frames
        :type flow: 2D array (R²)

    """

    def __init__(self, start_frame, end_frame, flow_calculator=None, spatial_interp=None, initial_flow=None):

        """

            :param start_frame: 2D array representation of the variable at the start of the temporal interpolation domain
            :type start_frame: array_like

            :param end_frame: 2D array representation of the variable at the end of the temporal interpolation
            :type end_frame: array_like

            :param flow_calculator: Callable that computes the optical flow between two frames / arrays
            :type flow_calculator: (R² x R²) -> R² mapping

            :param spatial_interp: Callable that generates a 2D interpolant of a frame array with normalized coordinates
            :type spatial_interp: R² -> (R² -> R)

            :returns: None
        """

        if spatial_interp is None:
            spatial_interp = RectBivariateSpline
        if flow_calculator is None:
            flow_calculator = FarnebackFlow()
        if start_frame.shape != end_frame.shape:
            raise ValueError(f"start_frame and end_frame must have the same shape; instead, they have shapes "
                             f"{start_frame.shape} and {end_frame.shape}")
        if len(start_frame.shape) != 2:
            raise ValueError(f"Input frames with shape {start_frame.shape} are not 2D")

        self.start_frame = start_frame
        self.end_frame = end_frame
        self.shape = self.start_frame.shape
        self.flow_calculator = flow_calculator
        self.spatial_interp = spatial_interp
        self.flow = self.flow_calculator(self.start_frame, self.end_frame, initial_flow)
        r_initial_flow = None if initial_flow is None else -initial_flow
        self.flow_r = - self.flow_calculator(self.end_frame, self.start_frame, r_initial_flow)
        self.I_start = NormalizedWrappedInterpolant(start_frame, spatial_interp)
        self.I_end = NormalizedWrappedInterpolant(end_frame, spatial_interp)
        self.ii, self.jj = np.meshgrid(self.I_start.i, self.I_start.j, indexing='ij')
        self.i_max = self.shape[0] - 1
        self.j_max = self.shape[1] - 1

    def __call__(self, t):

        """
            :param t: Fraction of the time domain between the start and the end frame. Must fulfill 0 <= t <= 1
            :type t: float

            :returns: Flow-interpolated array at t
            :rtype: array_like

        """

        if t < 0.0 or t > 1.0:
            raise ValueError("The temporal fraction for optical flow interpolation must lie between 0 and 1")

        ii_end = np.clip(self.ii + self.flow[:, :, 1] * (1 - t), 0, self.i_max)
        jj_end = np.clip(self.jj + self.flow[:, :, 0] * (1 - t), 0, self.j_max)
        ii_start = np.clip(self.ii - self.flow[:, :, 1] * t, 0, self.i_max)
        jj_start = np.clip(self.jj - self.flow[:, :, 0] * t, 0, self.j_max)

        start_translated = self.I_start(ii_start.flatten(), jj_start.flatten(), grid=False).reshape(self.shape)
        end_translated = self.I_end(ii_end.flatten(), jj_end.flatten(), grid=False).reshape(self.shape)

        return t * end_translated + (1 - t) * start_translated


class MultiFrameInterpolant(object):
    """Performs optical flow or linear interpolation between several consecutive, equally spaced frames

        :var interpolants: List of pairwise interpolants
        :type interpolants: List[TwoFrameInterpolant]

    """
    def __init__(self, frames, flow_calculator=None, spatial_interp=None, t_axis=None, initial_flows=None):
        """

            :param frames: List of 2D arrays
            :type frames: list[array_like]

            :param flow_calculator: Callable that computes the optical flow between two frames / arrays
            :type flow_calculator: (R² x R²) -> R² mapping

            :param spatial_interp: Callable that generates a 2D interpolant of a frame array with normalized coordinates
            :type spatial_interp: R² -> (R² -> R)

            :param t_axis: (optional) array containing the timestamps of t
            :type t_axis: array_like(dtype='timedelta64')

            :param initial_flows: List of arrays of shape N x M x 2 representing initial guesses for the optical flows
            :type initial_flows: list[array_like]

            :returns: None

        """
        assert len(frames) > 1
        if initial_flows is None:
            initial_flows = [None for _ in range(len(frames) - 1)]
        self.frames = frames
        self.t_axis = t_axis
        self.interpolants = [TwoFrameInterpolant(prev_frame, next_frame, flow_calculator, spatial_interp, flow)
                             for prev_frame, next_frame, flow in zip(frames[:-1], frames[1:], initial_flows)]

    def __call__(self, t):

        """

            :param t: Fraction of the time domain between the start and the end frame. Must fulfill 0 <= t <= n_frames
            :type t: float

            :returns: Flow-interpolated array at t
            :rtype: array_like

        """

        if t < 0.0 or t > len(self.frames) - 1:
            raise ValueError("The temporal fraction must lie between 0 and n - 1, where n is the number of frames")
        i = int(t)
        t_frac = t - i
        return self.interpolants[i](t_frac)

    def eval_at_t(self, t, check_nans=True):
        """

            :param t: Union[np.datetime64, np.timedelta64]
            :type t:

            :param check_nans: if True, replace any nans in the input with zeros
            :type check_nans: bool

            :returns:  Flow-interpolated array at timestamp t
            :rtype: array_like

        """

        upper_bound = self.t_axis[-1]
        lower_bound = self.t_axis[0]

        if self.t_axis is None:
            raise ValueError(f"The current {self.__class__} instance was not initialized with a temporal axis")
        if t < lower_bound:
            raise ValueError(f"The timestamp {t} is below the lower temporal axis bound of {lower_bound}")
        if t > upper_bound:
            raise ValueError(f"The timestamp {t} is below the upper temporal axis bound of {upper_bound}")

        idx_low, idx_high = get_bracketing_indexes(self.t_axis, t)
        t_high = self.t_axis[idx_high]
        t_low = self.t_axis[idx_low]
        frac = float(t - t_low) / float(t_high - t_low)
        assert 0 <= frac
        assert 1 >= frac
        output = self.interpolants[idx_low](frac)
        if check_nans:
            assert not np.isnan(output.sum())
        return output


class FarnebackFlow(object):
    """Optical flow calculator using the algorithm of Farneback (2003)

        :var parameters: arguments passed to the Farneback algorithm
        :type parameters: dict


    """

    def __init__(self, **kwargs):
        """Optical flow calculator using the algorithm of Farneback (2003)

            :param kwargs: arguments passed to the Farneback algorithm
            :type kwargs: dict

            :returns: None


        """
        self.parameters = {'iterations': 2,
                           'levels': 9,
                           'poly_n': 10,
                           'poly_sigma': 1.934,
                           'pyr_scale': 0.85,
                           'scale': 25.0,
                           'winsize': 77,
                           'flags': cv2.OPTFLOW_FARNEBACK_GAUSSIAN
                           }
        # self.parameters = {'pyr_scale': 0.5,
        #                    'levels': 5,
        #                    'winsize': 100,
        #                    'iterations': 4,
        #                    'poly_n': 9,
        #                    'poly_sigma': 1.6,
        #                    'flags': cv2.OPTFLOW_FARNEBACK_GAUSSIAN,
        #                    'scale': 1e5}
        self.parameters.update(kwargs)
        for key in ('levels', 'winsize', 'iterations', 'poly_n'):
            self.parameters[key] = int(self.parameters[key])
        self.scale = self.parameters['scale']
        del self.parameters['scale']

    def __call__(self, start_frame, end_frame, initial_flow=None, check_nans=True):
        """Returns the computed optical flow

            :param start_frame: 2D array (N x M) representing the start frame
            :type start_frame: array_like

            :param end_frame: 2D array (N x M) representing the end frame
            :type end_frame: array_like

            :param initial_flow: (optional) N x M x 2 array to initialize the flow
            :type initial_flow: array_like or None

            :param start_frame: 2D array (N x M) representing the start frame
            :type start_frame: array_like

            :param check_nans: (optional) if True, replace any nans in the input with zeros
            :type check_nans: bool

            :returns: optical flow between the frames
            :rtype: array_like


        """
        flow = cv2.calcOpticalFlowFarneback(self.scale * start_frame,
                                            self.scale * end_frame,
                                            initial_flow,
                                            **self.parameters)
        if check_nans:
            assert not np.isnan(flow.sum())
        return flow


class DummyFlow(object):
    def __init__(self):
        pass

    def __call__(self, start_frame, end_frame, initial_flow=None):
        return np.zeros(start_frame.shape + (2,))



# -*- coding: utf-8 -*-

import time
from collections import namedtuple
from typing import NamedTuple, List
from itertools import product
from metintos.optiflow import *
from metintos.constants import hour
from hyperopt import fmin, tpe, hp, STATUS_OK, Trials

#: :obj:`tuple` :
#: TimeSelector = namedtuple("TimeSelector", ['time', 'step'])
TimeSelector = namedtuple("TimeSelector", ['time', 'step'])


def add_times_to_steps_sources_targets(steps_sources_targets, dataset):
    """Given a dict mapping (source step, source step) -> (target step), return a dict replacing step keys and values
    with analogous TimeSelector values that also include time information.

        :param steps_sources_targets:
        :type steps_sources_targets:

        :param dataset:
        :type dataset:

        :returns: None

    """
    new_dict = {}
    times = dataset.coords['time'].values
    for t in times:
        for sources, target in steps_sources_targets.items():
            source_0 = TimeSelector(time=t, step=sources[0])
            source_1 = TimeSelector(time=t, step=sources[1])
            target = TimeSelector(time=t, step=target)
            new_dict[(source_0, source_1)] = target
    return new_dict


def add_steps_to_times_sources_targets(times_sources_targets, steps_list=(0*hour,)):
    """Given a dict mapping (source step, source step) -> (target step), return a dict replacing step keys and values
    with analogous TimeSelector values that also include time information.

        :param times_sources_targets:
        :type times_sources_targets:

        :param steps_list: default 0
        :type steps_list:

        :returns: None

    """
    new_dict = {}
    for s in steps_list:
        for sources, target in times_sources_targets.items():
            source_0 = TimeSelector(time=sources[0], step=s)
            source_1 = TimeSelector(time=sources[1], step=s)
            target = TimeSelector(time=target, step=s)
            new_dict[(source_0, source_1)] = target
    return new_dict


def filter_dims(d, dims):
    """Returns d filtered so that only key-value pairs where the key is in dims are kept

        :param d:
        :type d:

        :param dims:
        :type dims:

        :returns: None
    """
    return {k: v for k, v in d.items() if k in dims}


def evaluate_flow_on_dataset(flow_calc, dataset, variables, sources_and_targets,
                             array_norm=None, aggregation_norm=None, cut_evaluation_array=None,
                             aggregate=True):
    """

        :param flow_calc:
        :type flow_calc:

        :param dataset:
        :type dataset: xarray.Dataset

        :param variables:
        :type variables:

        :param sources_and_targets:
        :type sources_and_targets:

        :param array_norms:
        :type array_norms:

        :param aggregation_norm:
        :type aggregation_norm:

        :param cut_evaluation_array:
        :type cut_evaluation_array:

        :param aggregate:
        :type aggregate:

        :returns: None

    """
    if array_norm is None:
        def array_norm(x): return np.linalg.norm(x, ord=2)
    if aggregation_norm is None:
        def aggregation_norm(x): return np.linalg.norm(x, ord=2)
    non_ll_coords = {k: v for k, v in dataset.coords.items()
                     if k not in ('latitude', 'longitude', 'step', 'valid_time', 'time')
                     and v.values.size > 1}
    coord_names = non_ll_coords.keys()
    coord_axes = [non_ll_coords[c_name].values for c_name in coord_names]
    var_norms = {}
    for var in variables:
        norms_dict = {}
        norms = []
        for sources, target in sources_and_targets.items():
            sources: List[NamedTuple]
            source_selector_0 = filter_dims(sources[0]._asdict(), dataset.dims)
            source_selector_1 = filter_dims(sources[1]._asdict(), dataset.dims)
            target_selector = filter_dims(target._asdict(), dataset.dims)
            coord_values = product(*coord_axes)
            for cv in coord_values:
                selector = dict(zip(coord_names, cv))
                ds_subset = dataset[var].sel(**selector)
                source_start = ds_subset.loc[source_selector_0].values
                source_end = ds_subset.loc[source_selector_1].values
                target_frame = ds_subset.loc[target_selector].values
                interpolant = TwoFrameInterpolant(source_start, source_end, flow_calc)
                prediction_at_half_point = interpolant(.5)
                delta_array = prediction_at_half_point - target_frame
                if cut_evaluation_array is None:
                    norms.append(array_norm(delta_array))
                else:
                    cea = cut_evaluation_array
                    if aggregate:
                        norms.append(array_norm(delta_array[cea:-cea, cea:-cea]))
                    else:
                        norms_dict[(sources, target)] = array_norm(delta_array[cea:-cea, cea:-cea])
        if aggregate:
            var_norms[var] = aggregation_norm(norms)
        else:
            var_norms[var] = norms_dict
    return var_norms


def evaluate_flow_on_dataset_norms(flow_calc, dataset, variables, sources_and_targets,
                             array_norms=None, aggregation_norm=None, cut_evaluation_array=None,
                             aggregate=False):
    """

        :param flow_calc:
        :type flow_calc:

        :param dataset:
        :type dataset: xarray.Dataset

        :param variables:
        :type variables:

        :param sources_and_targets:
        :type sources_and_targets:

        :param array_norms:
        :type array_norms:

        :param aggregation_norm:
        :type aggregation_norm:

        :param cut_evaluation_array:
        :type cut_evaluation_array:

        :param aggregate:
        :type aggregate:

        :returns: None

    """
    if array_norms is None:
        def arr_norm(x): return np.linalg.norm(x, ord=2)
        array_norms = [arr_norm]
    if aggregation_norm is None:
        def aggregation_norm(x): return np.linalg.norm(x, ord=2)
    non_ll_coords = {k: v for k, v in dataset.coords.items()
                     if k not in ('latitude', 'longitude', 'step', 'valid_time', 'time')
                     and v.values.size > 1}
    coord_names = non_ll_coords.keys()
    coord_axes = [non_ll_coords[c_name].values for c_name in coord_names]
    var_norms = {}
    for var in variables:
        norms_dict = {arr_norm: {} for arr_norm in array_norms}
        norms = []
        for sources, target in sources_and_targets.items():
            sources: List[NamedTuple]
            source_selector_0 = filter_dims(sources[0]._asdict(), dataset.dims)
            source_selector_1 = filter_dims(sources[1]._asdict(), dataset.dims)
            target_selector = filter_dims(target._asdict(), dataset.dims)
            coord_values = product(*coord_axes)
            for cv in coord_values:
                selector = dict(zip(coord_names, cv))
                ds_subset = dataset[var].sel(**selector)
                source_start = ds_subset.loc[source_selector_0].values
                source_end = ds_subset.loc[source_selector_1].values
                target_frame = ds_subset.loc[target_selector].values
                interpolant = TwoFrameInterpolant(source_start, source_end, flow_calc)
                prediction_at_half_point = interpolant(.5)
                delta_array = prediction_at_half_point - target_frame
                if cut_evaluation_array is None:
                    # fix for disaggregated norms
                    norms.append(array_norm(delta_array))
                else:
                    cea = cut_evaluation_array
                    if aggregate:
                        for array_norm in array_norms:
                            norms_dict[array_norm][(sources, target)] = array_norm(delta_array[cea:-cea, cea:-cea])
                    else:
                        pass  # integrate later
        if aggregate:
            var_norms[var] = aggregation_norm(norms)
        else:
            var_norms[var] = norms_dict
    return var_norms


def optimize_flow(parameter_space, flow_generator, dataset, sources_and_targets,
                  var_weights=None, max_evals=100, cut_evaluation_array=None):
    """Optimizes the parameters of the algorithm for performance on the dataset

        :param parameter_space:
        :type parameter_space:

        :param flow_generator:
        :type flow_generator:

        :param dataset:
        :type dataset:

        :param sources_and_targets:
        :type sources_and_targets:

        :param var_weights:
        :type var_weights:

        :param max_evals:
        :type max_evals:

        :param cut_evaluation_array:
        :type cut_evaluation_array:

        :return: None


    """
    if var_weights is None:
        var_weights = {'u': 1, 'v': 1, 't': 1}
    variables = var_weights.keys()

    def objective(args):
        flow = flow_generator(args)
        t0 = time.time()
        losses = evaluate_flow_on_dataset(flow, dataset, variables, sources_and_targets,
                                          cut_evaluation_array=cut_evaluation_array)
        delta_t = time.time() - t0
        loss = sum(w * losses[var] for var, w in var_weights.items())
        print(loss, delta_t)
        print(args)
        print("===")
        return {'loss': loss,
                'status': STATUS_OK,
                'var_losses': losses,
                'time': delta_t}

    trials = Trials()
    best = fmin(objective,
                space=parameter_space,
                algo=tpe.suggest,
                max_evals=max_evals,
                trials=trials)

    return best, trials


if __name__ == "__main__":
    farneback_space = {
        'pyr_scale': hp.uniform('pyr_scale', 0.1, 0.9),
        'levels': hp.qloguniform('levels', 0, np.log(11), q=1),
        'winsize': hp.qlognormal('winsize', np.log(100), 1/6*np.log(100), 1),
        'iterations': hp.quniform('iterations', 2, 6, 1),
        'poly_n': hp.quniform('poly_n', 2, 14, 1),
        'poly_sigma': hp.uniform('poly_sigma', 0.1, 2.0),
        'scale': hp.qloguniform('scale', np.log(0.1), np.log(1e7), 1),
    }

    def flow_calculator(args):
        return FarnebackFlow(**args)

    snt_steps = {(0*hour, 2*hour): hour, (0*hour, 4*hour): 2*hour, (0*hour, 6*hour): 3*hour}





#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from setuptools import setup
from os import path
from io import open

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
  name = 'metintos',         # How you named your package folder (MyLib)
  packages = ['metintos'],   # Chose the same as "name"
  version = '0.2.2',      # Start with a small number and increase it with every change you make
  license='lgpl-3.0',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'Meteorological Interpolation Toolbox for Optimization and Simulation',   # Give a short description about your library
  long_description=long_description,
  long_description_content_type='text/markdown',
  author = 'Daniel González Arribas',                   # Type in your name
  author_email = 'dangonza@ing.uc3m.es',      # Type in your E-Mail
  url = 'https://github.com/javiergarciaheras/mitos',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/javiergarciaheras/mitos/archive/v_01.tar.gz',    # I explain this later on
  keywords = ['Optical flow', 'Meteorological', 'Interpolation', 'Optimization', 'Simulation'],   # Keywords that define your package best
  install_requires=[            # I get to this in a second
          'xarray',
          'numpy',
          'casadi',
          'scipy',
          'hyperopt',
          'setuptools',
          'opencv-python'
      ],
  extras_require={
          'tuning': ['hyperopt'],
      },
  include_package_data=True,
  tests_require=["pytest"],
  zip_safe=False,
  classifiers=[ # Chose the classifiers from here: https://pypi.org/classifiers/
    'Development Status :: 4 - Beta',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Science/Research',      # Define that your audience are developers
    'Intended Audience :: Education',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',   # Again, pick a license
    'Operating System :: MacOS :: MacOS X',
    'Operating System :: Microsoft :: Windows',
    'Operating System :: Unix',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.5', #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
  ],
)


#!/usr/bin/env python
# -*- coding: utf-8 -*-

import numpy as np
# Unless specified otherwise, all units are SI

#: :obj:`float` :
#: kappa = 1.4
kappa = 1.4

#: :obj:`float` :
#: Rg = 287.05287
Rg = 287.05287

#: :obj:`float` :
#: ft = 0.3048
ft = 0.3048

#: :obj:`float` :
#: m2ft = ft**-1
m2ft = ft**-1

#: :obj:`float` :
#: T0_MSL = 288.15
T0_MSL = 288.15

#: :obj:`float` :
#: P0_MSL = 101325.0
P0_MSL = 101325.0

#: :obj:`float` :
#: rho0_MSL = 1.225
rho0_MSL = 1.225

#: :obj:`float` :
#: a0_MSL = 340.294
a0_MSL = 340.294

#: :obj:`float` :
#: nmi = 1852.0
nmi = 1852.0

#: :obj:`float` :
#: kt = nmi/3600
kt = nmi/3600

#: :obj:`float` :
#: ms2kt = 1/kt
ms2kt = 1/kt

#: :obj:`float` :
#: g0 = g = 9.80665
g0 = g = 9.80665

#: :obj:`float` :
#: lapse_rate_troposphere = -0.0065
lapse_rate_troposphere = -0.0065

#: :obj:`float` :
#: R_mean = 6356.8e3
R_mean = 6356.8e3

#: :obj:`float` :
#: d2r = np.pi/180
d2r = np.pi/180

#: :obj:`float` :
#: flattening = 1/298.257223563
flattening = 1/298.257223563

#: :obj:`float` :
#: first_eccentricity = (flattening*(2 - flattening))**0.5
first_eccentricity = (flattening*(2 - flattening))**0.5

#: :obj:`float` :
#: R_a = 6378.137e3
R_a = 6378.137e3

#: :obj:`float` :
#: R_b = 6356.752
R_b = 6356.752

#: :obj:`float` :
#: h_tropopause = 11e3
h_tropopause = 11e3

#: :obj:`float` :
#: T_tropopause = 216.65
T_tropopause = 216.65

#: :obj:`float` :
#: minute = np.timedelta64(1, 'm').astype('timedelta64[ns]')
minute = np.timedelta64(1, 'm').astype('timedelta64[ns]')

#: :obj:`float` :
#: hour = np.timedelta64(1, 'h').astype('timedelta64[ns]')
hour = np.timedelta64(1, 'h').astype('timedelta64[ns]')
